#version 430
#extension GL_ARB_compute_shader: enable
//#extension GL_ARB_shader_storage_buffer_object: enable;

layout(location = 8) uniform float elapsedTime;
layout(location = 9) uniform int maxParticles;
layout(location = 10) uniform vec3 direction;
layout(location = 11) uniform float speed;
layout(binding = 1, offset = 0) uniform atomic_uint Count;

layout(std430, binding=2) buffer positionIn {
	vec4 positionsIn[];
};
layout(std430, binding=3) buffer velocityIn{
	vec4 velocitiesIn[];
};
layout(std430, binding=4) buffer positionOut {
	vec4 positionsOut[];
};
layout(std430, binding=5) buffer velocityOut{
	vec4 velocitiesOut[];
};
layout( local_size_x = 128,  local_size_y = 1, local_size_z = 1 ) in;

void main(){
	uint id = gl_GlobalInvocationID.x;
	vec4 pos = positionsIn[id];
	vec4 velo = velocitiesIn[id];
	pos.w = pos.w - elapsedTime;//w = TTL

	if(pos.w > 1.f) {
		//the first particle does not move backwards
		uint nr = atomicCounterIncrement(Count);
		positionsOut[nr]=pos;
		velocitiesOut[nr]=velocitiesIn[id];

		//... and should spawn new ones every time
		nr = atomicCounterIncrement(Count);
		if(nr < maxParticles){
			positionsOut[nr]=pos;
			positionsOut[nr].w = 1.f;
			velocitiesOut[nr]=vec4(-direction*speed, 0.f);
			nr = atomicCounterIncrement(Count);
			if(nr < maxParticles){
				positionsOut[nr]=pos;
				positionsOut[nr].w = 1.f;
				velocitiesOut[nr]=vec4(-direction*speed/2, 0.f);
			} else {
				atomicCounterDecrement(Count);
			}
		} else {
			atomicCounterDecrement(Count);
		}
	} else if(pos.w > 0) {
		uint nr = atomicCounterIncrement(Count);
		positionsOut[nr]=vec4(pos.xyz + velo.xyz*elapsedTime, pos.w);
		velocitiesOut[nr]=velo;
	}
}