
vec3 calculateDiffuse(vec3 lightVector, vec3 lightColor) {
    float cosTheta = clamp(dot(normalize(normal_world),lightVector), 0,1 ); // = n dot l
    return (kd * textureColorDiffuse() * lightColor * cosTheta);
}

vec3 calculateSpecular(vec3 lightVector, vec3 viewVector, vec3 lightColor) {
    vec3 r = reflect(-lightVector, normalize(normal_world));
    float cosPhi = clamp(dot(viewVector, r), 0,1 );
    return (ks * textureColorSpecular() * pow(cosPhi,alpha) * lightColor);
}

vec3 calculatePointLight(PointLight point, vec3 viewVector){
    vec3 lPoint = point.position - position_world;
    float dist = length(lPoint);
    lPoint = normalize(lPoint);
    float fAtt = 1/(point.attenuationConstant + dist * point.attenuationLinear + dist * dist * point.attenuationQuadratic);

    return calculateDiffuse(lPoint, point.color) * fAtt 
        + calculateSpecular(lPoint, viewVector, point.color) * fAtt;
}

float pcfShadowTest(vec4 fragmentLightPosition, vec3 lSpot){

    vec3 fragmentLightPosition_0_1 = (fragmentLightPosition.xyz/fragmentLightPosition.w)*0.5+0.5;//transform from [-w,w] to [0,1]
    float antiAcneCreme = max(0.05 * (1.0 - dot(normal_world, lSpot)), 0.005);  //shadow bias but it does fight shadow acne, so...
    
    float result = 0.f;
    vec2 texelSize = 1.0 / textureSize(shadowMap, 0);

    for(int u = -1; u <= 1; u++)
        for(int v = -1; v <= 1; v++)
        {
            float lightDepth = texture(shadowMap, fragmentLightPosition_0_1.xy+vec2(u,v)*texelSize).r;
            bool inShadow = lightDepth < fragmentLightPosition_0_1.z-antiAcneCreme 
                && fragmentLightPosition_0_1.z <= 1.0f; //to make sure that things behind the light's far plane arent considered in shadow

            if(inShadow)
                result = result + 1.f;
        }
    return 1-result/9.f;
}

vec3 calculateSpotLight(SpotLight spot, vec3 viewVector, vec4 fragmentLightPosition){
    vec3 lSpot = spot.position - position_world;
    float dist = length(lSpot);
    lSpot = normalize(lSpot);

    float cosAngleInCone = dot(normalize(-spot.direction), lSpot);
    float angleInCone = acos(cosAngleInCone);

    float shadowFactor = pcfShadowTest(fragmentLightPosition, lSpot);

    if(angleInCone < spot.angleOuter && shadowFactor > 0.f){
            
        float falloff = 1;
        if(angleInCone >= spot.angleInner){
            falloff = 1-smoothstep(spot.angleInner, spot.angleOuter, angleInCone);
        }
        
        float attenuationConstant = spot.attenuation.x;
        float attenuationLinear = spot.attenuation.y;
        float attenuationQuadratic = spot.attenuation.z;

        float fAttSpot = 1/(attenuationConstant + dist * attenuationLinear + dist * dist * attenuationQuadratic);

        vec3 light = calculateDiffuse(lSpot, spot.color) * fAttSpot * falloff
            + calculateSpecular(lSpot, viewVector, spot.color) * fAttSpot * falloff;
        return light * shadowFactor;
    }
    return vec3(0);
}

vec3 calculateDirectional(DirectionalLight directionalLight, vec3 viewVector){
    vec3 lDirectional = normalize(-directionalLight.direction);
    return calculateDiffuse(lDirectional, directionalLight.color)
        + calculateSpecular(lDirectional, viewVector, directionalLight.color);
 }

vec4 calculateLights(vec3 viewVector){
        vec3 ambient = ambientPower * 0.5f * ka * textureColorDiffuse();
        vec3 result = ambient;

        //directional
        for(int i = 0; i < numDirectional; i++)
            result += calculateDirectional(directionals[i], viewVector);
        
        //point
        for(int i = 0; i < numPoint; i++)
            result += calculatePointLight(pointLights[i], viewVector);

        //spot
        for(int i = 0; i < numSpot; i++)
            result += calculateSpotLight(spotLights[i], viewVector, position_light[i]);

        return vec4(result, 0);
}
